from __future__ import annotations

import json
from typing import Any, Optional, Type, cast

from pydantic import BaseModel
from rcheck import r

from pnorm import MarshallRecordException, ParamType, T


def get_params(
    name: str, params: Optional[ParamType], by_alias: bool = False
) -> dict[str, Any]:
    if params is None:
        return {}

    if isinstance(params, BaseModel):
        params = params.model_dump(by_alias=by_alias)

    return cast(dict[str, Any], r.check_mapping(name, params, keys_of=str))


def combine_into_return(
    return_model: Type[T],
    result: dict[str, Any] | BaseModel,
    params: Optional[ParamType] = None,
) -> T:
    result_dict = get_params("Query Result", result)

    if params is not None:
        result_dict.update(get_params("Query Params", params))

    try:
        return return_model(**result_dict)
    except Exception as e:
        model_name = getattr(return_model, "__name__")
        record = json.dumps(result_dict)
        msg = f"Could not marshall record {record} into model {model_name}"
        raise MarshallRecordException(msg) from e



from pydantic import AliasChoices, BaseModel, ConfigDict, Field


class PostgresCredentials(BaseModel):
    dbname: str = Field(
        default="postgres",
        validation_alias=AliasChoices("dbname", "database"),
    )
    user: str
    password: str
    host: str
    port: int = 5432

    model_config = ConfigDict(extra="forbid")



from __future__ import annotations

from typing import Any, Mapping, TypeVar

from pydantic import BaseModel

T = TypeVar("T", bound=BaseModel)

QueryParams = Mapping[str, Any]
ParamType = QueryParams | BaseModel



from __future__ import annotations

from contextlib import contextmanager
from types import TracebackType
from typing import TYPE_CHECKING, Generator, Optional, Type

if TYPE_CHECKING:
    from pnorm import PostgresClient


# todo: rm
class Session:
    def __init__(self, client: PostgresClient):
        self.client = client
        self.original_auto_create_connection = self.client.auto_create_connection
        self.client.auto_create_connection = False

    def __enter__(self) -> PostgresClient:
        if self.client.connection is None:
            self.client.create_connection()

        return self.client

    def __exit__(
        self,
        exc_type: Optional[Type[Exception]],
        exc_value: Optional[Exception],
        exc_tb: Optional[TracebackType],
    ):
        if self.client.connection is not None:
            if exc_type is not None:
                self.client.rollback()

            self.client.close_connection()

        self.client.auto_create_connection = self.original_auto_create_connection


@contextmanager
def create_session(client: PostgresClient) -> Generator[None, None, None]:
    original_auto_create_connection = client.auto_create_connection
    client.auto_create_connection = False
    close_connection_after_use = False

    if client.connection is None:
        client.create_connection()
        close_connection_after_use = True

    try:
        yield
    except:
        client.rollback()
        raise
    finally:
        if client.connection is not None and close_connection_after_use:
            client.close_connection()

        client.auto_create_connection = original_auto_create_connection


@contextmanager
def create_transaction(client: PostgresClient) -> Generator[None, None, None]:
    client.start_transaction()

    try:
        yield
    except:
        client.rollback()
        raise
    finally:
        client.end_transaction()



from .credentials import PostgresCredentials  # type: ignore
from .exceptions import *
from .types import *

...  # type: ignore

from .client import PostgresClient
from .contexts import *

# https://github.com/dagster-io/dagster/blob/master/python_modules/libraries/dagster-aws/dagster_aws/redshift/resources.py
# https://github.com/jmoiron/sqlx
# https://jmoiron.github.io/sqlx/



import inspect
import os
import sys
from datetime import datetime
from glob import glob
from importlib import import_module, reload
from pathlib import Path
from typing import Literal, Type, cast

import yaml
from pydantic import BaseModel

from pnorm import PostgresClient, PostgresCredentials, Session, create_transaction


class BaseUsers(BaseModel):
    username: str


class VersionChange(BaseModel):
    from_version: int | None
    to_version: int
    update_type: Literal["upgrade"] | Literal["downgrade"]
    updated_at: datetime
    description: str


class Migration:
    def __init__(
        self,
        client: PostgresClient,
        version_number: int,
        version_description: str | None = None,
    ):
        self.client = client
        self.version_number = version_number
        self.version_description = version_description

    def upgrade(self, from_version: int | None):
        description = self.version_description

        if description is None:
            description = f"Upgrading from {from_version} to {self.version_number}"

        self.client.execute(
            "insert into version_log (from_version, to_version, description, update_type, updated_at) values (%(from_version)s, %(to_version)s, %(description)s, %(update_type)s, %(updated_at)s)",
            VersionChange(
                from_version=from_version,
                to_version=self.version_number,
                update_type="upgrade",
                updated_at=datetime.now(),
                description=description,
            ),
        )

    def downgrade(self, to_version: int):
        self.client.execute(
            "insert into version_log (from_version, to_version, description, update_type, updated_at) values (%(from_version)s, %(to_version)s, %(description)s, %(update_type)s, %(updated_at)s)",
            VersionChange(
                from_version=self.version_number,
                to_version=to_version,
                update_type="downgrade",
                updated_at=datetime.now(),
                description=f"Downgrading from {self.version_number} to {to_version}",
            ),
        )

    def run(self, current_version: VersionChange):
        if current_version.to_version > self.version_number:
            self.downgrade(current_version.to_version)
        elif current_version.to_version < self.version_number:
            self.upgrade(current_version.to_version)


class MigrationConfig(BaseModel):
    version: int
    description: str


class MigrationMetadata(BaseModel):
    path: Path
    config: MigrationConfig


class Exists(BaseModel):
    exists: bool


def base_migration(client: PostgresClient):
    exists_check = client.get(
        Exists,
        "select exists ("
        "  select from information_schema.tables"
        "  where table_name = 'version_log'"
        ")",
    )

    if exists_check is not None and exists_check.exists:
        return

    with create_transaction(client):
        client.execute(
            "create table version_log ("
            "  from_version integer null"
            "  , to_version integer null"
            "  , description varchar"
            "  , update_type varchar not null"
            "  , updated_at timestamp not null"
            ")"
        )

        timestamp = datetime.now()

        client.execute(
            "insert into version_log values ("
            "  null"
            "  , 0"
            "  , 'Create version table'"
            "  , 'upgrade'"
            f"  , '{timestamp}'"
            ")"
        )


def load_migration(migration: MigrationMetadata) -> Type[Migration]:
    # sys.path.append(str(file_path.parent.parent))
    old_path = sys.path

    sys.path = [str(migration.path)]

    module = import_module("migration", migration.path.name)
    module = reload(module)
    classes = module.__dict__.values()

    output = None

    for cls in classes:
        if inspect.isclass(cls) and issubclass(cls, Migration) and cls != Migration:
            print("LOADING MODULE", cls.__name__, "FROM", migration.path)
            # sys.path.pop()
            # sys.path = old_path
            # return cls

            output = cls

    if output is not None:
        return output

    sys.path = old_path
    raise Exception()


def main(folder_path: Path, credentials: PostgresCredentials):
    migrations: list[MigrationMetadata] = []

    with open(
        folder_path / "config.yaml",
        "r",
        encoding="utf-8",
    ) as file:
        file_contents = yaml.safe_load(file)

    expected_version = file_contents["version"]

    for migration_path in glob("*", root_dir=folder_path):
        if not os.path.isdir(folder_path / migration_path):
            continue

        with open(
            folder_path / migration_path / "migration.yaml",
            "r",
            encoding="utf-8",
        ) as file:
            file_contents = yaml.safe_load(file)

        migrations.append(
            MigrationMetadata(
                path=folder_path / migration_path,
                config=MigrationConfig(**file_contents),
            )
        )

    with Session(PostgresClient(credentials)) as session:
        base_migration(session)

        current_version = session.get(
            VersionChange,
            "select * from version_log order by updated_at desc limit 1",
        )

        if current_version is None:
            raise Exception()

    print("CURRENT VERSION", current_version, "\n\n------------")
    previous_version = current_version.to_version

    if expected_version == previous_version:
        return
    elif expected_version > previous_version:
        versions_to_upgrade = [
            m
            for m in migrations
            if m.config.version > previous_version
            and m.config.version <= expected_version
        ]
        versions_to_upgrade = sorted(
            versions_to_upgrade, key=lambda x: x.config.version
        )

        print("VERSIONS TO UPGRADE", versions_to_upgrade)

        for version in versions_to_upgrade:
            print("from ", previous_version, "to ", version.config.version)
            migration = load_migration(version)
            migration(
                session, version.config.version, version.config.description
            ).upgrade(previous_version)
            previous_version = version.config.version
    elif expected_version < previous_version:
        print(migrations, expected_version, previous_version)
        versions_to_upgrade = [
            m
            for m in migrations
            if m.config.version <= previous_version
            and m.config.version >= expected_version
        ]
        versions_to_upgrade = sorted(
            versions_to_upgrade, key=lambda x: -x.config.version
        )
        print("------")
        print(versions_to_upgrade)

        for version, next_version in zip(
            versions_to_upgrade[:-1], versions_to_upgrade[1:]
        ):
            migration = load_migration(version)
            migration(
                session, version.config.version, version.config.description
            ).downgrade(next_version.config.version)

        if expected_version == 0:
            version = versions_to_upgrade[-1]
            migration = load_migration(version)
            migration(
                session, version.config.version, version.config.description
            ).downgrade(0)
